package com.steerinfo.inPlantNavigation.service.impl;

import com.alibaba.fastjson.JSONObject;
import com.steerinfo.inPlantNavigation.exception.VertexAngEdgeException;
import com.steerinfo.inPlantNavigation.mapper.MapVertexMapper;
import com.steerinfo.inPlantNavigation.model.IPMMSVertex;
import com.steerinfo.inPlantNavigation.model.IPMMSVertexEdge;
import com.steerinfo.inPlantNavigation.model.MapVertexEdge;
import com.steerinfo.inPlantNavigation.util.ExcelIOUtil;
import com.steerinfo.inPlantNavigation.model.MapVertex;
import com.steerinfo.inPlantNavigation.service.IMapVertexService;
import org.apache.commons.lang.SerializationUtils;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.*;
import java.math.BigDecimal;

/**
 * MapVertex服务实现：
 * @author generator
 * @version 1.0-SNAPSHORT 2022-04-27 06:36
 * 类描述
 * 修订历史：
 * 日期：2022-04-27
 * 作者：generator
 * 参考：
 * 描述：MapVertex服务实现
 * @see null
 * @Copyright 湖南视拓信息技术股份有限公司. All rights reserved.
 */
@Service(value = "mapVertexService")
public class MapVertexServiceImpl implements IMapVertexService {

    @Autowired
    private MapVertexMapper mapVertexMapper;


    @Override
    public int importVertex(String filePath) throws IOException {
        //清空表，再添加
        filePath="C://Users//liyunguang//Desktop//达钢测绘地点坐标数据_new.xls";
        List<MapVertex> vertexList=new ArrayList<>();
        //现在使用的是自己手动在地图上查询到的坐标
        List<HSSFRow> rows= ExcelIOUtil.getRowsByXls(filePath,2);
        System.out.println("读取Excel完成，共"+rows.size()+"行");
        for(HSSFRow row: rows){
            if(row.getCell(0)==null || row.getCell(0).equals("")){
                continue;
            }
            MapVertex temp=new MapVertex();
            try{
                temp.setVertexId(new BigDecimal(row.getCell(0).toString()));//标点id
            }catch (NumberFormatException e){
                System.out.println(row.getCell(0).toString()+"转换成Number失败,跳过这行");
                continue;
            }
            temp.setAddressName(row.getCell(1).toString());//地名
            String []lonlat=row.getCell(10).toString().split(",");//切割字符串
            temp.setLongitude(new BigDecimal(lonlat[0]));//经度
            temp.setLatitude(new BigDecimal(lonlat[1]));//纬度
            temp.setStatus(new BigDecimal(0));
            System.out.println(temp.toString());
            vertexList.add(temp);
        }
        return mapVertexMapper.batchInsert(vertexList);
    }

    @Override
    public List<MapVertex> findAllAvailable() {
        return mapVertexMapper.findAllAvailable();
    }

    /**
     * 返回门岗、汽车衡和作业点
     * @return
     */
    @Override
    public ArrayList<MapVertex> findSelections() {
        ArrayList<MapVertex> list=mapVertexMapper.findSelections();//查询的结果包含重复值，需要去重
        ArrayList<MapVertex> selections=new ArrayList<>();
        HashSet<BigDecimal> set=new HashSet<>();
        for(MapVertex vertex:list){
            if(!set.contains(vertex.getVertexId())){
                set.add(vertex.getVertexId());
                selections.add(vertex);
            }
            else{
                continue;
            }
        }
        System.out.println(selections);
        return selections;
    }

    @Override
    public Map<String, MapVertex> initMapVertex() {
        List<MapVertex> list=mapVertexMapper.findAllAvailable();
        Map<String,MapVertex> vertexs=new HashMap<>();
        for (MapVertex vertex:list){
            vertexs.put(vertex.getVertexId().toString(),vertex);
        }
        return vertexs;
    }

    @Override
    public ArrayList<MapVertex> getObtainTheOptimalPath(String startpoint, String endPoint, Map<String, List<MapVertexEdge>> vertexEdgeList, Map<String, MapVertex> vertexList) throws VertexAngEdgeException {
        //初始化数据
        Integer[] startPoin=new Integer[]{0,-1,0};
        //创建S战队（放入的是已经遍历过边的顶点）
        List<String> obtainTheOptimalPath=new ArrayList();
        //obtainTheOptimalPath.add(statpoint);
        //创建 dist[]  (起点到集合点中的最短路径) key：指向的点 new int[]{0,-1,0}; {累计权重、中间点、自己权重}
        HashMap<String,Integer[]> currentBestDinstance=new HashMap<>();
        currentBestDinstance.put(startpoint,startPoin);
        while (obtainTheOptimalPath.size()!=vertexList.size()){
            // dist[]最小值
            String stringMinimumValue = getStringMinimumValue(currentBestDinstance, obtainTheOptimalPath);

            //加入S战队
            obtainTheOptimalPath.add(stringMinimumValue);

            //判断是否有进入的边
            if (island(currentBestDinstance,stringMinimumValue,vertexEdgeList)){
                continue;
            }

            //借东风
            List<MapVertexEdge> ipmmsVertexEdges = vertexEdgeList.get(stringMinimumValue);

            //判断是否存在边
            if (ipmmsVertexEdges==null){
                continue;
            }

            //遍历所有的边
            for(MapVertexEdge ipmmsVertexEdge :ipmmsVertexEdges){
                //箭头出发点
                BigDecimal outVertexID = ipmmsVertexEdge.getOutvertexId();
                //到该点的最短路径
                Integer[] starBestPath = currentBestDinstance.get(outVertexID.toString());
                //被指向的顶点
                BigDecimal inVertexID= ipmmsVertexEdge.getInvertexId();
                Integer[] historyBestPaht = currentBestDinstance.get(inVertexID.toString());
                if (starBestPath==null){
                    throw new VertexAngEdgeException("这条边不存在！"+outVertexID.toString()+"->"+inVertexID.toString());
                }else  if (!outVertexID.toString().equals(stringMinimumValue)){
                    throw new VertexAngEdgeException(outVertexID+"->"+inVertexID+"不是顶点"+stringMinimumValue+"的边！");
                }
                //判断是否存在其路线到这个点的距离。如果不存在则将这条线路作为起点到该点最短路径、如果本条路线是最短路径需要替换最短路径
                if(historyBestPaht==null){
                    Integer distance=starBestPath[0]+ipmmsVertexEdge.getWeigh().intValue();
                    Integer outVertex=outVertexID.intValue();
                    Integer[] bestPath=new Integer[]{distance,outVertex,ipmmsVertexEdge.getWeigh().intValue()};
                    currentBestDinstance.put(inVertexID.toString(),bestPath);
                }else if ((starBestPath[0]+ipmmsVertexEdge.getWeigh().intValue())<historyBestPaht[0]){
                    //更新和此顶点有关的数据
                    Integer distance=starBestPath[0]+ipmmsVertexEdge.getWeigh().intValue();
                    Integer outVertex=outVertexID.intValue();
                    Integer[] bestPath=new Integer[]{distance,outVertex,ipmmsVertexEdge.getWeigh().intValue()};
                    currentBestDinstance.put(inVertexID.toString(),bestPath);
                    ArrayList<String> updateList = new ArrayList<>();
                    updateRelactionVertex(updateList, stringMinimumValue, currentBestDinstance);
                }
            }
        }
        if (!currentBestDinstance.containsKey(endPoint)){
            throw new VertexAngEdgeException("没有该条路径，请检查是否缺少边！"+endPoint);
        }
        return startPointToEndPointPath(endPoint,currentBestDinstance,vertexList);
    }


    //获得目前最短距离
    public String getStringMinimumValue(HashMap<String,Integer[]> currentBestDinstance,List<String> obtainTheOptimalPath){

        HashMap<String,Integer[]> bestDinstance = (HashMap<String, Integer[]>) SerializationUtils.clone(currentBestDinstance);

        for (String item:obtainTheOptimalPath){
            if (bestDinstance.containsKey(item)){
                bestDinstance.remove(item);
            }
        }
        Set<String> keys = bestDinstance.keySet();
        String vertex="";
        int minimumValue=10000000;
        for (String key :keys){
            Integer[] value = currentBestDinstance.get(key);
            if (minimumValue>value[0]){
                minimumValue=value[0];
                vertex=key;
            }
        }
        return vertex;
    }

    //判断是否有进入这个顶点的边，如果没有这个顶点就是一个孤岛
    public Boolean island(HashMap<String,Integer[]> currentBestDinstance,String stringMinimumValue,Map<String, List<MapVertexEdge>> vertexEdgeList){
        if(!currentBestDinstance.containsKey(stringMinimumValue)){
            Collection<List<MapVertexEdge>> values = vertexEdgeList.values();
            for (List<MapVertexEdge> vertexEdges : values){
                for (MapVertexEdge vertexEdge:vertexEdges){
                    if (vertexEdge.getInvertexId().equals(stringMinimumValue))
                        return false;
                }
            }
            return  true;
        }
        return false;
    }

    //更新相关的值
    public void updateRelactionVertex(List<String> result,String currentVertex,HashMap<String,Integer[]> currentBestDinstance){
        Set<String> keys = currentBestDinstance.keySet();
        for (String  key:  keys){
            Integer[] value=currentBestDinstance.get(key);
            if (new BigDecimal(value[1]).equals(new BigDecimal(currentVertex))){
                value[0]=value[2]+currentBestDinstance.get(currentVertex)[0];
                currentBestDinstance.put(key,value);
                updateRelactionVertex(result,key,currentBestDinstance);
            }
        }

    }
    //通过所有最优解集合选择我们要的路径
    public ArrayList<MapVertex> startPointToEndPointPath(String point, HashMap<String,Integer[]> currentBestDinstance, Map<String, MapVertex> vertexList){
        ArrayList<MapVertex> obtainOptimalPath=new ArrayList();
        Integer beforeDistance=0;
        do {
            Integer[] ints = currentBestDinstance.get(point);
            beforeDistance=ints[1];
            obtainOptimalPath.add(vertexList.get(point));
            point=String.valueOf(ints[1]);
        }
        while (beforeDistance!=-1);
        return obtainOptimalPath;
    }

    /**
     * 取得订单路径细节
     * @param orderId
     * @return
     */
    @Override
    public Map<String, Object> findOrderDetails(String orderId) {
        Map<String, Object> result=mapVertexMapper.findOrderDetails(orderId);
        return result;
    }

    /**
     * 根据订单id和前后步骤取得目标路径
     * @param orderId
     * @param startStep
     * @param endStep
     * @param vertexList
     * @param vertexEdgeList
     * @return
     * @throws VertexAngEdgeException
     */
    @Override
    public ArrayList<MapVertex> getPathByOrderID(String orderId,String startStep,String endStep,Map<String, MapVertex> vertexList,Map<String, List<MapVertexEdge>> vertexEdgeList) throws VertexAngEdgeException {
        //校验
        if(orderId==null || startStep==null ||endStep==null){
            return null;
        }
        //初始化
        ArrayList<MapVertex> obtainOptimalPath=null;
        Map<String, Object> details=mapVertexMapper.findOrderDetails(orderId);
        String startPoint=null;
        String endPoint=null;
        //取得起点的顶点
        startPoint=parseVertexID(startStep,details);
        //取得终点的顶点
        endPoint=parseVertexID(endStep,details);
        //不为空则找寻最短路径
        if(startPoint!=null && endPoint!=null){
            obtainOptimalPath= getObtainTheOptimalPath(startPoint,endPoint,vertexEdgeList,vertexList);
            Collections.reverse(obtainOptimalPath);//翻转
        }
        return obtainOptimalPath;
    }
    //转换成顶点编号
    private String parseVertexID(String step,Map<String, Object> details){
        //校验数据
        if(step==null || details==null || details.size()<1){
            return null;
        }
        //转换成顶点id
        String point=null;
        switch (step){
            case "进厂":
                if(details.get("entryGatepostId")!=null)
                    point=mapVertexMapper.getVertexIdByGatePostId(details.get("entryGatepostId").toString());
                break;
            case "计皮":
                if(details.get("tareCalculateId")!=null)
                    point=mapVertexMapper.getVertexIdByCalculateId(details.get("tareCalculateId").toString());
                break;
            case "装货":
                if(details.get("loadingID")!=null)
                    point=mapVertexMapper.getVertexIdByWarehouseId(details.get("loadingID").toString());
                break;
            case "卸货":
                if(details.get("unloadWarhouseId")!=null)
                    point=mapVertexMapper.getVertexIdByWarehouseId(details.get("unloadWarhouseId").toString());
                break;
            case "计毛":
                if(details.get("grossPlaceId")!=null)
                    point=mapVertexMapper.getVertexIdByCalculateId(details.get("grossPlaceId").toString());
                break;
            case "出厂":
                if(details.get("leaveGatepostId")!=null)
                    point=mapVertexMapper.getVertexIdByGatePostId(details.get("leaveGatepostId").toString());
                break;
        }
        return point;
    }
}
